from .fields import DjangoConnectionField
from .optimizer import optimize, required_fields
from .types import DjangoObjectType

__all__ = [
    "DjangoConnectionField",
    "DjangoObjectType",
    "optimize",
    "required_fields",
]



import graphene_django
from django.db.models import Model, QuerySet

from .optimizer import optimize
from .settings import optimizer_settings
from .typing import PK, GQLInfo, Optional, TypeVar
from .utils import can_optimize

TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "DjangoObjectType",
]


class DjangoObjectType(graphene_django.types.DjangoObjectType):
    """DjangoObjectType that automatically optimizes its queryset."""

    class Meta:
        abstract = True

    @classmethod
    def max_complexity(cls) -> int:
        return optimizer_settings.MAX_COMPLEXITY  # type: ignore[no-any-return]

    @classmethod
    def get_queryset(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        if can_optimize(info):
            queryset = optimize(queryset, info, cls.max_complexity())
        return queryset

    @classmethod
    def get_node(cls, info: GQLInfo, id: PK) -> Optional[TModel]:  # noqa: A002
        queryset: QuerySet[TModel] = cls._meta.model.objects.filter(pk=id)
        if can_optimize(info):
            queryset = optimize(queryset, info, cls.max_complexity(), pk=id)
        return queryset.first()



import graphene_django.filter
from django.db.models import Model, QuerySet
from django.db.models.manager import Manager
from graphene.relay.connection import Connection
from graphql_relay import EdgeType
from graphql_relay.connection.connection import ConnectionType

from .cache import store_in_query_cache
from .optimizer import QueryOptimizer
from .typing import Any, Callable, GQLInfo, Optional, TypeAlias, TypeVar
from .utils import get_field_type, get_selections

TModel = TypeVar("TModel", bound=Model)

__all__ = [
    "ConnectionFieldCachingMixin",
    "DjangoConnectionField",
]

Args: TypeAlias = tuple[
    Callable[..., Optional[Manager[TModel]]],  # resolver
    Connection,  # connection
    Manager[TModel],  # default_manager
    Callable[..., QuerySet[TModel]],  # queryset_resolver
    int,  # max_limit
    bool,  # enforce_first_or_last
    Optional[Model],  # enforce_first_or_last
    GQLInfo,  # info
]


class ConnectionFieldCachingMixin:
    """Mixin to add query caching to connection fields."""

    @classmethod
    def connection_resolver(cls, *args: Any, **kwargs: Any) -> ConnectionType:
        args: Args  # type: ignore[no-redef]
        connection_field: graphene_django.fields.DjangoConnectionField = super()  # type: ignore[override]
        connection_type: ConnectionType = connection_field.connection_resolver(*args, **kwargs)
        cache_edges(edges=connection_type.edges, info=args[7])
        return connection_type


class DjangoConnectionField(
    ConnectionFieldCachingMixin,
    graphene_django.fields.DjangoConnectionField,
):
    pass


def cache_edges(edges: list[EdgeType], info: GQLInfo) -> None:
    """
    Cache edges received from a connection in the query cache.

    :param edges: Edges containing the fetched models.
    :param info: The GraphQLResolveInfo object used in the optimization process.
    """
    if not edges:
        return

    field_type = get_field_type(info)
    selections = get_selections(info)
    optimizer = QueryOptimizer(info)
    store = optimizer.optimize_selections(field_type, selections, model=type(edges[0].node))
    store_in_query_cache(
        key=info.operation,
        items=(edge.node for edge in edges),
        schema=info.schema,
        store=store,
    )



from django.core.exceptions import FieldDoesNotExist
from django.db.models import ForeignKey, ManyToOneRel, Model, QuerySet
from graphene.relay.connection import ConnectionOptions
from graphene.types.definitions import GrapheneObjectType, GrapheneUnionType
from graphene.utils.str_converters import to_snake_case
from graphene_django.types import DjangoObjectTypeOptions
from graphql import (
    FieldNode,
    FragmentSpreadNode,
    GraphQLField,
    GraphQLOutputType,
    InlineFragmentNode,
    SelectionNode,
)

from .cache import get_from_query_cache, store_in_query_cache
from .settings import optimizer_settings
from .store import QueryOptimizerStore
from .typing import (
    PK,
    Callable,
    GQLInfo,
    Iterable,
    ModelField,
    Optional,
    ToManyField,
    ToOneField,
    TypeOptions,
    TypeVar,
    Union,
)
from .utils import (
    get_field_type,
    get_selections,
    get_underlying_type,
    is_foreign_key_id,
    is_to_many,
    is_to_one,
)

TModel = TypeVar("TModel", bound=Model)
TCallable = TypeVar("TCallable", bound=Callable)


__all__ = [
    "optimize",
    "QueryOptimizer",
    "required_fields",
]


def optimize(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    max_complexity: Optional[int] = None,
    pk: PK = None,
) -> QuerySet[TModel]:
    """
    Optimize the given queryset according to the field selections
    received in the GraphQLResolveInfo.

    :param queryset: Base queryset to optimize from.
    :param info: The GraphQLResolveInfo object used in the optimization process.
    :param max_complexity: How many 'select_related' and 'prefetch_related' table joins are allowed.
                           Used to protect from malicious queries.
    :param pk: Primary key for an item in the queryset model. If set, optimizer will check
               the query cache for that primary key before making query.
    :return: The optimized queryset.
    """
    # Check if prior optimization has been done already
    if queryset._hints.get(optimizer_settings.OPTIMIZER_MARK, False):  # type: ignore[attr-defined]
        return queryset

    field_type = get_field_type(info)
    selections = get_selections(info)
    if not selections:  # pragma: no cover
        return queryset

    optimizer = QueryOptimizer(info)
    store = optimizer.optimize_selections(field_type, selections, queryset.model)

    max_complexity = max_complexity if max_complexity is not None else optimizer_settings.MAX_COMPLEXITY
    complexity = store.complexity
    if complexity > max_complexity:
        msg = f"Query complexity of {complexity} exceeds the maximum allowed of {max_complexity}"
        raise RuntimeError(msg)

    if pk is not None:
        cached_item = get_from_query_cache(info.operation, info.schema, queryset.model, pk, store)
        if cached_item is not None:
            queryset._result_cache = [cached_item]
            return queryset

    queryset = store.optimize_queryset(queryset, pk=pk)
    if optimizer.cache_results:
        store_in_query_cache(info.operation, queryset, info.schema, store)

    return queryset


def required_fields(*fields: str) -> Callable[[TCallable], TCallable]:
    """
    Annotate custom field resolver to require given fields
    in relation to its DjangoObjectType model.

    :param fields: Fields that the decorated resolver needs.
                   Related entity fields can also be used with
                   the field lookup syntax (e.g., 'related__field')
    """

    def decorator(resolver: TCallable) -> TCallable:
        resolver.hints = fields  # type: ignore[attr-defined]
        return resolver

    return decorator


class QueryOptimizer:
    """Query optimizer for Django QuerySets."""

    def __init__(self, info: GQLInfo) -> None:
        self.info = info
        self.cache_results = True

    def optimize_selections(
        self,
        field_type: Union[GrapheneObjectType, GrapheneUnionType],
        selections: tuple[SelectionNode, ...],
        model: type[Model],
    ) -> QueryOptimizerStore:
        store = QueryOptimizerStore(model=model)

        for selection in selections:
            if isinstance(selection, FieldNode):
                self.optimize_field_node(field_type, selection, store)

            elif isinstance(selection, FragmentSpreadNode):
                self.optimize_fragment_spread(field_type, selection, model, store)

            elif isinstance(selection, InlineFragmentNode):
                self.optimize_inline_fragment(field_type, selection, model, store)

            else:  # pragma: no cover
                msg = f"Unhandled selection node: '{selection}'"
                raise TypeError(msg)

        return store

    def optimize_field_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        store: QueryOptimizerStore,
    ) -> None:
        options: TypeOptions = field_type.graphene_type._meta

        if isinstance(options, ConnectionOptions):
            return self.handle_connection_node(field_type, selection, store)

        if isinstance(options, DjangoObjectTypeOptions):
            return self.handle_regular_node(field_type, selection, store)

        msg = f"Unhandled field options type: {options}"  # pragma: no cover
        raise TypeError(msg)  # pragma: no cover

    def handle_regular_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        store: QueryOptimizerStore,
    ) -> None:
        model: type[Model] = field_type.graphene_type._meta.model
        selection_graphql_name = selection.name.value
        selection_graphql_field = field_type.fields.get(selection_graphql_name)
        if selection_graphql_field is None:  # pragma: no cover
            return

        model_field_name = to_snake_case(selection_graphql_name)
        try:
            model_field: ModelField = model._meta.get_field(model_field_name)
        except FieldDoesNotExist:
            self.add_hinted_fields(selection_graphql_field, model, store)
            return

        if not model_field.is_relation or is_foreign_key_id(model_field, model_field_name):
            store.only_fields.append(model_field_name)

        elif is_to_one(model_field):  # noinspection PyTypeChecker
            self.handle_to_one(model_field_name, selection, selection_graphql_field.type, model_field, store)

        elif is_to_many(model_field):  # noinspection PyTypeChecker
            self.handle_to_many(model_field_name, selection, selection_graphql_field.type, model_field, store)

        else:  # pragma: no cover
            msg = f"Unhandled selection: '{selection.name.value}'"
            raise TypeError(msg)

    def add_hinted_fields(self, field: GraphQLField, model: type[Model], store: QueryOptimizerStore) -> None:
        hints: Optional[tuple[str, ...]] = getattr(field.resolve, "hints", None)
        if hints is None:  # pragma: no cover
            return

        model_fields: list[ModelField] = model._meta.get_fields()
        for field_name in hints:
            hint_store = QueryOptimizerStore(model=model)
            self.find_field_from_model(field_name, model_fields, hint_store)
            store += hint_store

    def find_field_from_model(
        self,
        field_name: str,
        model_fields: Iterable[ModelField],
        store: QueryOptimizerStore,
        prefix: str = "",
    ) -> None:
        for model_field in model_fields:
            model_field_name = model_field.name
            if prefix:
                model_field_name = f"{prefix}__{model_field_name}"

            if field_name == model_field_name:
                store.only_fields.append(model_field.name)
                return None

            if field_name.startswith(model_field_name):
                related_model: Optional[type[Model]] = model_field.related_model
                if related_model is None:  # pragma: no cover
                    msg = f"No related model, but hint seems like has one: {field_name!r}"
                    raise ValueError(msg)

                nested_store = QueryOptimizerStore(model=related_model)
                if is_to_many(model_field):
                    store.prefetch_stores[model_field.name] = (nested_store, related_model.objects.all())
                elif is_to_one(model_field):
                    store.select_stores[model_field.name] = nested_store
                else:  # pragma: no cover
                    msg = f"Field {model_field} is not a related field."
                    raise TypeError(msg)

                if isinstance(model_field, ManyToOneRel):  # Add connecting entity
                    nested_store.only_fields.append(model_field.field.name)

                related_model_fields: list[ModelField] = related_model._meta.get_fields()

                return self.find_field_from_model(
                    field_name=field_name,
                    prefix=model_field_name,
                    store=nested_store,
                    model_fields=related_model_fields,
                )

        msg = f"Field {field_name!r} not found in fields: {model_fields}."  # pragma: no cover
        raise ValueError(msg)  # pragma: no cover

    def handle_connection_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        store: QueryOptimizerStore,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        # Connection QuerySets are sliced, so the results
        # should be cached later in the connection field.
        self.cache_results = False

        node: FieldNode = selection.selection_set.selections[0]  # type: ignore[assignment]
        if node.selection_set is None:  # page info
            return

        edges_field = field_type.fields[selection.name.value]
        edge_type = get_underlying_type(edges_field.type)
        node_field = edge_type.fields[node.name.value]
        node_type = get_underlying_type(node_field.type)
        node_model: type[Model] = node_type.graphene_type._meta.model

        nested_store = self.optimize_selections(node_type, node.selection_set.selections, node_model)
        store += nested_store

    def handle_to_one(
        self,
        model_field_name: str,
        selection: FieldNode,
        selection_field_type: GraphQLOutputType,
        model_field: ToOneField,
        store: QueryOptimizerStore,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        selection_field_type = get_underlying_type(selection_field_type)
        nested_store = self.optimize_selections(
            selection_field_type,
            selection.selection_set.selections,
            model_field.model,
        )

        if isinstance(model_field, ForeignKey):  # Add connecting entity
            store.only_fields.append(model_field_name)

        store.select_stores[model_field_name] = nested_store

    def handle_to_many(
        self,
        model_field_name: str,
        selection: FieldNode,
        selection_field_type: GraphQLOutputType,
        model_field: ToManyField,
        store: QueryOptimizerStore,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        selection_field_type = get_underlying_type(selection_field_type)
        nested_store = self.optimize_selections(
            selection_field_type,
            selection.selection_set.selections,
            model_field.model,
        )

        if isinstance(model_field, ManyToOneRel):  # Add connecting entity
            nested_store.only_fields.append(model_field.field.name)

        related_queryset: QuerySet[Model] = model_field.related_model.objects.all()
        store.prefetch_stores[model_field_name] = nested_store, related_queryset

    def optimize_fragment_spread(
        self,
        field_type: GrapheneObjectType,
        selection: FragmentSpreadNode,
        model: type[Model],
        store: QueryOptimizerStore,
    ) -> None:
        graphql_name = selection.name.value
        field_node = self.info.fragments[graphql_name]
        selections = field_node.selection_set.selections
        nested_store = self.optimize_selections(field_type, selections, model)
        store += nested_store

    def optimize_inline_fragment(
        self,
        field_type: GrapheneUnionType,
        selection: InlineFragmentNode,
        model: type[Model],
        store: QueryOptimizerStore,
    ) -> None:
        fragment_type_name = selection.type_condition.name.value
        selection_graphql_field: Optional[GrapheneObjectType]
        selection_graphql_field = next((t for t in field_type.types if t.name == fragment_type_name), None)
        if selection_graphql_field is None:  # pragma: no cover
            return

        fragment_model: type[Model] = selection_graphql_field.graphene_type._meta.model
        if fragment_model != model:  # pragma: no cover
            return

        selections = selection.selection_set.selections
        nested_store = self.optimize_selections(selection_graphql_field, selections, fragment_model)
        store += nested_store



from dataclasses import dataclass

from django.db.models import Model, Prefetch, QuerySet
from django.db.models.constants import LOOKUP_SEP

from .settings import optimizer_settings
from .typing import PK, TypeVar
from .utils import unique

TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "QueryOptimizerStore",
]


@dataclass
class CompilationResults:
    only_fields: list[str]
    select_related: list[str]
    prefetch_related: list[Prefetch]


class QueryOptimizerStore:
    """Store for holding optimization data."""

    def __init__(self, model: type[Model]) -> None:
        self.model = model
        self.only_fields: list[str] = []
        self.select_stores: dict[str, "QueryOptimizerStore"] = {}
        self.prefetch_stores: dict[str, tuple["QueryOptimizerStore", QuerySet[Model]]] = {}

    def compile(self, *, in_prefetch: bool = False) -> CompilationResults:  # noqa: A003
        results = CompilationResults(
            only_fields=self.only_fields.copy(),
            select_related=[],
            prefetch_related=[],
        )

        for name, store in self.select_stores.items():
            if not in_prefetch:
                results.select_related.append(name)
            self._compile_nested(name, store, results, in_prefetch=in_prefetch)

        for name, (store, queryset) in self.prefetch_stores.items():
            optimized_queryset = store.optimize_queryset(queryset)
            results.prefetch_related.append(Prefetch(name, optimized_queryset))
            self._compile_nested(name, store, results, in_prefetch=True)

        results.only_fields = unique(results.only_fields)
        results.select_related = unique(results.select_related)
        results.prefetch_related = unique(results.prefetch_related)
        return results

    @staticmethod
    def _compile_nested(
        name: str,
        store: "QueryOptimizerStore",
        results: CompilationResults,
        *,
        in_prefetch: bool,
    ) -> None:
        nested_results = store.compile(in_prefetch=in_prefetch)
        if not in_prefetch:
            results.only_fields.extend(f"{name}{LOOKUP_SEP}{only}" for only in nested_results.only_fields)

        results.select_related.extend(f"{name}{LOOKUP_SEP}{select}" for select in nested_results.select_related)
        for prefetch in nested_results.prefetch_related:
            prefetch.add_prefix(name)
            results.prefetch_related.append(prefetch)

    def optimize_queryset(
        self,
        queryset: QuerySet[TModel],
        *,
        pk: PK = None,
    ) -> QuerySet[TModel]:
        results = self.compile()

        if results.prefetch_related:
            queryset = queryset.prefetch_related(*results.prefetch_related)
        if results.select_related:
            queryset = queryset.select_related(*results.select_related)
        if results.only_fields and not optimizer_settings.DISABLE_ONLY_FIELDS_OPTIMIZATION:
            queryset = queryset.only(*results.only_fields)
        if pk is not None:
            queryset = queryset.filter(pk=pk)

        # Mark queryset as optimized so that later optimizers know to skip optimization
        queryset._hints[optimizer_settings.OPTIMIZER_MARK] = True  # type: ignore[attr-defined]

        return queryset

    @property
    def complexity(self) -> int:
        value: int = 0
        for store in self.select_stores.values():
            value += store.complexity
        for store, _ in self.prefetch_stores.values():
            value += store.complexity
        return value + len(self.select_stores) + len(self.prefetch_stores)

    def __add__(self, other: "QueryOptimizerStore") -> "QueryOptimizerStore":
        self.only_fields += other.only_fields
        self.select_stores.update(other.select_stores)
        self.prefetch_stores.update(other.prefetch_stores)
        return self

    def __str__(self) -> str:
        results = self.compile()
        only = ",".join(results.only_fields)
        select = ",".join(results.select_related)
        prefetch = ",".join(item.prefetch_to for item in results.prefetch_related)
        return f"{only=}|{select=}|{prefetch=}"



import graphene_django.filter

from .fields import ConnectionFieldCachingMixin

__all__ = [
    "DjangoFilterConnectionField",
]


class DjangoFilterConnectionField(
    ConnectionFieldCachingMixin,
    graphene_django.filter.DjangoFilterConnectionField,
):
    pass



from collections import defaultdict
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from django.db.models import Model, QuerySet
from graphql import GraphQLSchema

from .settings import optimizer_settings
from .typing import PK, Hashable, Iterable, Optional, QueryCache, TableName, TypeVar

if TYPE_CHECKING:
    from .store import QueryOptimizerStore


TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "get_from_query_cache",
    "get_query_cache",
    "store_in_query_cache",
]


def get_query_cache(key: Hashable, schema: GraphQLSchema) -> QueryCache:
    """
    Get or create a cache for storing model instances.
    Cache is implemented as a WeakKeyDictionary on the given key,
    stored in the given schema object.

    Items in the cache are stored first by their database table, then by
    their selected fields (as determined by QueryOptimizerStore), and lastly
    by their primary key. The first two levels of the hierarchy are implemented
    as defaultdicts.

    :param key: Any hashable value that is present only for the duration of
                a single request, e.g., 'info.operation'.
    :param schema: The GraphQLSchema object where the cache will exist.
    :return: The cache.
    """
    cache = schema.extensions.setdefault(optimizer_settings.QUERY_CACHE_KEY, WeakKeyDictionary())
    return cache.setdefault(key, defaultdict(lambda: defaultdict(dict)))  # type: ignore[no-any-return]


def get_from_query_cache(
    key: Hashable,
    schema: GraphQLSchema,
    model: type[TModel],
    pk: PK,
    store: "QueryOptimizerStore",
) -> Optional[TModel]:
    """
    Get the given model instance from query cache.

    :param key: Any hashable value that is present only for the duration of
                a single request, e.g., 'info.operation'.
    :param schema: The GraphQLSchema object where the cache exists.
    :param model: The model type to look for.
    :param pk: The primary key of the model instance to look for.
    :param store: The QueryOptimizerStore describing the fields that
                  should have been fetched on the model instance.
    :return: The Model instance if it exists in the cache, None if not.
    """
    store_str = str(store)
    query_cache = get_query_cache(key, schema)
    return query_cache[model._meta.db_table][store_str].get(pk)


def store_in_query_cache(
    key: Hashable,
    items: Iterable[Model],
    schema: GraphQLSchema,
    store: "QueryOptimizerStore",
) -> None:
    """
    Set all given models, as well as any related models joined to them
    as described by the given QueryOptimizerStore, to the query cache.

    :param key: Any hashable value that is present only for the duration of
                a single request, e.g., 'info.operation'.
    :param items: Model instances that should be stored in the query cache.
    :param schema: The GraphQLSchema object where the cache exists.
    :param store: The QueryOptimizerStore describing the fields that
                  are fetched on the model instances.
    """
    query_cache = get_query_cache(key, schema)
    items = list(items)  # For QuerySets, the database query will occur here
    annotations = _get_annotations(items[0])
    for item in items:
        _add_item(query_cache, item, annotations, store)


def _get_annotations(item: Model) -> list[str]:
    # Don't use 'queryset.query.annotations' since we cannot extract annotations
    # as cleanly from the model if some other iterable is given.
    # (these results contain foreign key ids as well)
    model_builtins = {"_prefetched_objects_cache", "_state"}
    fields: set[str] = {field.name for field in item._meta.get_fields()}
    attributes: set[str] = set(item.__dict__)
    diff = attributes.difference(fields)
    return list(diff.difference(model_builtins))


def _add_item(query_cache: QueryCache, instance: Model, annotations: list[str], store: "QueryOptimizerStore") -> None:
    store_str = str(store)
    if annotations:
        store_str += f"|{annotations=}"
    table_name: TableName = instance._meta.db_table
    query_cache[table_name][store_str][instance.pk] = instance
    _add_selected(query_cache, instance, store)
    _add_prefetched(query_cache, instance, store)


def _add_selected(query_cache: QueryCache, instance: Model, store: "QueryOptimizerStore") -> None:
    for nested_name, nested_store in store.select_stores.items():
        nested_instance: Model = getattr(instance, nested_name)
        _add_item(query_cache, nested_instance, [], nested_store)


def _add_prefetched(query_cache: QueryCache, instance: Model, store: "QueryOptimizerStore") -> None:
    for nested_name, (nested_store, _queryset) in store.prefetch_stores.items():
        selected: QuerySet[Model] = getattr(instance, nested_name).all()
        for select in selected:
            _add_item(query_cache, select, [], nested_store)



from django.db.models import ForeignKey
from graphene import Connection
from graphene.types.definitions import GrapheneObjectType
from graphene_django import DjangoObjectType
from graphql import GraphQLOutputType, SelectionNode
from graphql.execution.execute import get_field_def

from .typing import Collection, GQLInfo, ModelField, ToManyField, ToOneField, TypeGuard, TypeVar

__all__ = [
    "get_field_type",
    "get_selections",
    "get_underlying_type",
    "is_foreign_key_id",
    "is_to_many",
    "is_to_one",
    "unique",
]


T = TypeVar("T")


def is_foreign_key_id(model_field: ModelField, name: str) -> bool:
    return isinstance(model_field, ForeignKey) and model_field.name != name and model_field.get_attname() == name


def get_underlying_type(field_type: GraphQLOutputType) -> GrapheneObjectType:
    while hasattr(field_type, "of_type"):
        field_type = field_type.of_type
    return field_type


def is_to_many(model_field: ModelField) -> TypeGuard[ToManyField]:
    return bool(model_field.one_to_many or model_field.many_to_many)


def is_to_one(model_field: ModelField) -> TypeGuard[ToOneField]:
    return bool(model_field.many_to_one or model_field.one_to_one)


def get_field_type(info: GQLInfo) -> GrapheneObjectType:
    field_node = info.field_nodes[0]
    field_def = get_field_def(info.schema, info.parent_type, field_node)
    return get_underlying_type(field_def.type)


def get_selections(info: GQLInfo) -> tuple[SelectionNode, ...]:
    field_node = info.field_nodes[0]
    selection_set = field_node.selection_set
    return () if selection_set is None else selection_set.selections


def unique(items: Collection[T]) -> list[T]:
    return list(dict.fromkeys(items))


def can_optimize(info: GQLInfo) -> bool:
    return_type = get_underlying_type(info.return_type)

    return isinstance(return_type, GrapheneObjectType) and (
        issubclass(return_type.graphene_type, (DjangoObjectType, Connection))
    )



from typing import Any, Callable, Collection, Hashable, Iterable, NamedTuple, Optional, TypeVar, Union

from graphene.relay.connection import ConnectionOptions
from graphene_django.types import DjangoObjectTypeOptions

# New in version 3.10
try:
    from typing import TypeAlias, TypeGuard
except ImportError:
    from typing_extensions import TypeAlias, TypeGuard


import graphql
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import (
    Field,
    ForeignKey,
    ForeignObject,
    ForeignObjectRel,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    Model,
    OneToOneField,
)

__all__ = [
    "Any",
    "Callable",
    "Collection",
    "GQLInfo",
    "Hashable",
    "Iterable",
    "ModelField",
    "NamedTuple",
    "Optional",
    "PK",
    "QueryCache",
    "StoreStr",
    "TableName",
    "ToManyField",
    "ToOneField",
    "TypeGuard",
    "TypeOptions",
    "TypeVar",
    "Union",
]


class GQLInfo(graphql.GraphQLResolveInfo):
    context: WSGIRequest


TModel = TypeVar("TModel", bound=Model)
TableName: TypeAlias = str
StoreStr: TypeAlias = str
PK: TypeAlias = Any
QueryCache: TypeAlias = dict[TableName, dict[StoreStr, dict[PK, TModel]]]
ModelField: TypeAlias = Union[Field, ForeignObjectRel, GenericForeignKey]
ToManyField: TypeAlias = Union[GenericRelation, ManyToManyField, ManyToOneRel, ManyToManyRel]
ToOneField: TypeAlias = Union[GenericRelation, ForeignObject, ForeignKey, OneToOneField]
TypeOptions: TypeAlias = Union[DjangoObjectTypeOptions, ConnectionOptions]
